prefix=/usr/local
exec_prefix=${prefix}
libdir=${exec_prefix}/lib

PATH=${exec_prefix}/sbin:$PATH
export PATH

GLUSTERD_PIDFILEDIR=/var/run/gluster
export GLUSTERD_PIDFILEDIR

LD_LIBRARY_PATH=${exec_prefix}/lib:$LD_LIBRARY_PATH
export LD_LIBRARY_PATH

LIBRARY_PATH=${exec_prefix}/lib:$LIBRARY_PATH
export LIBRARY_PATH

CPATH=${prefix}/include:$CPATH
export CPATH

GLUSTERD_WORKDIR=/var/lib/glusterd
export GLUSTERD_WORKDIR

PKG_CONFIG_PATH=${libdir}/pkgconfig:$PKG_CONFIG_PATH
export PKG_CONFIG_PATH

PYTHONPATH=${prefix}/lib/python2.6/site-packages:$PYTHON_PATH
export PYTHONPATH

PYTHON=/usr/bin/python
export PYTHON

PYTHONPATH=${prefix}/lib/python2.6/site-packages:$PYTHON_PATH
export PYTHONPATH
